""" Functions for new scRNA data processing """
import os
import pandas as pd
import numpy as np
import scanpy as sc
import anndata
from typing import Optional, Union

###  function  ###



def isRawCount( ds ):
    """ 
    Tell whether the matrix is raw  count or float 
        if top 100 cells all mod 0, then determine it's raw count = integer

    Parameters
    ----------
    ds
        loom, or just a numpy matrix 2d. 
    
    Returns
    ----------
    int
        0, not raw count.
        1, is raw count.

    """
    if( ds[:100,:100].sum() - int( ds[:100,:100].sum() ) == 0 ) :
        datatype = 'rawcount'
        datatype = 1
    else:
        datatype = 'notcount'
        datatype = 0
    return(datatype)

### pre-processing adata

def adatapp(
    adata_input : anndata.AnnData, 
    mode: str = 'pp', # or  'umap'
    random_state = 42,
    MAX_MITO_PERCENT=5,
    MINGENES =500,
    MINCOUNTS = 1000,
    MINCELLS=5,
):

    """ 
    Pre-processing. 

    Parameters
    ----------
    adata_input
        scanpy scRNA object
        if adata.X is integer , or adata has 'counts' layer, like from 10X data.
        go through standard preprocessing.
        remove cells with <500 genes and <1000 counts. 
        remove genes with min cells < 5. 
        keep mitocondrial <5%. 
        normalize to 1e6. log. pca, umap. normalized value saved to 'norm_data' layer

        if adata.X is continuous, such as Smartseq data
        remove cells with <500 genes. remove genes with <5 cells. 
    mode
        choose from ['pp' , 'umap']
        pp: full pre-processing
        umap: preprocessing was done. only compute highly_varaible genes, pca, neighbors, umap.
    random_state
        umap random state
    MAX_MITO_PERCENT
    MINGENES 
    MINCOUNTS
    MINCELLS

    Returns
    ----------
        processed scanpy object

    Examples
    ----------
    >>> adata = adatapp(adata )

    """

    adata = adata_input.copy()

    if(mode=='pp'):
        isCount = isRawCount( adata.X )
        if( isCount ): 
            adata.layers["counts"] = adata.X.copy()
            print( '[Log] adata.X is count. Copy to .layers["counts"] . Pre-processing... ')
            adata.var_names_make_unique()
            adata.obs['n_counts'] = adata.X.sum(1).A1
            adata.obs['log_counts'] = np.log(adata.obs['n_counts']) #
            adata.obs['n_genes'] = (adata.X > 0).sum(1)
            print( '[Log] filtering ')
            sc.pp.filter_cells(adata, min_genes= MINGENES ) 
            sc.pp.filter_cells(adata, min_counts = MINCOUNTS )
            sc.pp.filter_genes(adata, min_cells = MINCELLS) 
            print('[Log] Human MT-percentage')
            adata.var['mt'] = adata.var_names.str.startswith('MT-') 
            sc.pp.calculate_qc_metrics(adata, qc_vars=['mt'], percent_top=None, log1p=False, inplace=True)
            adata = adata[adata.obs.pct_counts_mt < MAX_MITO_PERCENT, :]
            sc.pp.normalize_total(adata, target_sum = 1e4)  # match NormalizeData in Seurat
            sc.pp.log1p(adata) # checked the values are logged.
            sc.pp.highly_variable_genes(adata, flavor='cell_ranger', n_top_genes=4000 )
            sc.tl.pca(adata,  svd_solver='arpack')
            sc.pp.neighbors(adata )
            sc.tl.umap(adata, random_state = random_state)
            
        else: 
            # directly use the values if X not continuous: already normalized
            print( '[Log] adata.X is continuous . Pre-processing... ')
            adata.var_names_make_unique()
            adata.layers["norm_data"] = adata.X
            sc.pp.filter_cells(adata, min_genes=500) 
            sc.pp.filter_genes(adata, min_cells=5) 
            sc.pp.highly_variable_genes(adata, flavor='cell_ranger', n_top_genes=4000 )
            sc.tl.pca(adata,  svd_solver='arpack')
            sc.pp.neighbors(adata )
            sc.tl.umap(adata, random_state = random_state)

    elif(mode=='umap'):
        
        sc.pp.highly_variable_genes(adata, flavor='cell_ranger', n_top_genes=4000) 
        sc.tl.pca(adata)
        sc.pp.neighbors(adata)
        sc.tl.umap(adata, random_state = random_state)
    return(adata)


def load(
    file, 
    preprocessing = False,
    delimiter = ',',
):
    """ 
    Read in a expression matrix table file as scanpy object.

    If converting from an Seurat object, table can be saved in R:
        tmpdata = SeuratObject@assays$RNA@data
        write.csv( t( as.matrix(tmpdata) ), file = 'data.csv')

    Parameters
    ----------
    file
        file path for the expression file.
        if suffix is .h5 or .h5ad, data is loaded using scanpy.
        if not, load the file in format of a table text file
    preprocessing
        whether to do preprocessing for the loaded data
    delimiter
        delimiter of the expression matrix table, if file is a text table
    
    Returns
    ----------
    adata
        loaded scRNA data in scanpy object
        
    Examples
    ----------
    >>> adata = load( file, preprocessing=False )
    
    """
    adata = None

    suffix = file.split('.')[-1]
    if(suffix == 'h5' or suffix == 'h5ad'):
        annfile = file
        adata = sc.read(annfile)
    else:
        matfile = file
        adata = anndata.read_csv( matfile, delimiter = delimiter, first_column_names = True)
    if(not adata):
        return(None)

    if ( preprocessing == True ):
        #if 'norm_data' not in adata.layers:
        ## do preprocessing from raw
        print( ' Redo preprocessing ' )
        adata = adatapp( adata )
    else:
        print( ' Loaded ' + file)
    return(adata)



def batchharmonize(
    adata , 
    batchcols = [],
    random_state= 0
):
    """
    Harmonize batches such as patient, sample, using Harmony. Re-calculated neighbors, umap .

    Parameters
    ----------
    adata:
        scanpy object. 
    batchcols
        batch columns in adata.obs
    random_state
        umap random_state
    Returns
    ----------
    adata with batch correction and neighbors,umap calcualted

    """
    

    if((batchcols is not None) or (len(batchcols))>=1):
        print(f'[Log ] hamonize with batch ', batchcols)
        if('X_pca' not in adata.obsm.keys()):
            sc.tl.pca(adata,  svd_solver='arpack')
        try:
            sc.external.pp.harmony_integrate(adata, key = batchcols )
            sc.pp.neighbors(adata, use_rep = 'X_pca_harmony')
        except:
            print('[Log ] harmonization problem, skipping batch correction')
            sc.pp.neighbors(adata)
    else:
        print(f'[Log ]  no batch label detected. Computing neighbors using default')
        sc.pp.neighbors(adata)

    print(f'[Log ] Recomputing umap ' )
    sc.tl.umap(adata, random_state = random_state)
    return(adata)



"""
def isIndexNumberButNotGeneName( adata ):

    if( adata.var_names[0]=='0' ):
        return(True)
    else:
        return(False)
"""


def add_extra_metainfo( adata, meta ):
    """
    Add special extra metainformation for sample datasets

    Parameters
    ----------
    adata: anndata.AnnData
        scanpy object with scRNA expression
    meta: pd.DataFrame
        Dataframe with columns of metadata for each cell. index is cell barcode same as in adata.obs.index.
    
    Returns
    ----------
    anndata.AnnData
        scanpy object with new columns merged in obs. 

    Examples
    ----------
    >>> adata = testdata.add_extra_metainfo( adata, meta )
    """
    # add warning
    adata.obs = adata.obs.join( meta[[col for col in meta.columns if col not in adata.obs.columns]] )
    return(adata)


import pandas as pd
import os
import scanpy as sc
import numpy as np
import warnings
import anndata



############ project single cell to mec for creating new adata obj ###########

def projectMec(expr, mec, glstcol = 'TopGene20'):

    """ 
    Parameters
    ----------
    expr
        expression matrix, cell by gene. 
    mec
        mec table with one column, two types of format accepted
        format 1, each row is a mec with genes , comma seperated.
        format 2, each row is a genes, each column is a mec, value is float
    glstcol
        Used only when mec is the list format format1, and glstcol is the column name to record comma separated gene list for each mec.

    Returns
    ----------
    scorepd, cell by signature.

    Examples
    ---------
    >>> scorepd = mecmapper.projectMec( df, mec )
    
    """
    score_lst = []
   
    if(isinstance( mec.iloc[0,0], str) ) :
        # the mec format is format 1. no weighting.
        for i in mec.index:
            mec_i = str(i) 
            gs = mec.loc[i, glstcol ].split(',')
            gs_sub = [ t for t in expr.columns if t in gs ]
            score = pd.DataFrame( expr[ gs_sub ].mean(axis=1), columns= [ mec_i ] )
            score_lst.append( score  )
        scorepd = pd.concat(score_lst , axis=1).fillna(0)
    else:
        # the mec format is not format 1. using values as weight.
        shared_genes = mec.index.intersection( expr.columns )
        module_sharedgenes = mec.loc[ shared_genes ]
        expr_sharedgenes = expr[ shared_genes ]
        scorearray = expr_sharedgenes.values.dot( module_sharedgenes.values )
        scorepd = pd.DataFrame( scorearray, index = expr.index, columns = mec.columns)
        
    return( scorepd )



def annToDataFrame( adata_input, genescaling = False,  layer = 'norm_data'):
    """ 
    Extract expression matrix from scanpy object.
    Parameters
    ----------
    adata_input: anndata.AnnData
        Input scanpy object. 
    genescaling: bool
        Whether to z-scale extracted feature matrix.
    layer: str
        Layer of expression to extract from adata_input

    Returns
    ----------
    pd.dataframe
        Extracted expression matrix from scanpy object.
        
    """
    adata = adata_input.copy()
    try:
        df = adata.to_df(layer = layer )
    except:
        warnings.warn('Warning: no '+layer +' layer. using X.')
        df = adata.to_df()
    if(  genescaling ):
        df = (df - df.mean())/df.std()
    
    return( df )


def scale(df):
    """ standardize scaling feature """
    return(  (df-df.mean())/df.std()  )

def projectMecAnn( adata_input, 
                mec, 
                genescaling = False,  
                sigscaling = True, 
                addon = False, 
                layer = 'norm_data', 
                glstcol = 'TopGene20'):
    """
    Project single cell expression in AnnData to MeCs
    Calls: projectMeC,annToDataFrame

    Parameters
    ----------
    adata_input: anndata.AnnData
        Input scanpy object for gene expression
    mec: pd.DataFrame
        mec table with one column, two types of format both accepted
        format 1, each row is a mec with genes , comma seperated.
        format 2, each row is a genes, each column is a mec, value is float.
    genescaling: bool
        Whether to scale expression on gene level. Recommended to be False.
    sigscaling: bool
        Whether to scale projected scores across cells. Recomended and default is True.
    addon: bool
        Whether to keep the original adata and append the signautres in obs, or return an independent anndata (pdata) with only projected values (which saves memory).
    layer: str
        Layer of expression to extract from adata_input
    glstcol
        Used only when mec is the list format format 1, and glstcol is the column name to record comma separated gene list for each mec.


    Returns
    ----------
    anndata.AnnData
        If addon is False, return pdata where values are MeC-projected values. 
        If addon is True, return adata where values are same as in adata_input, but with extra obs columns.
    Examples
    ----------
    >>> pdata = mecmapper.projectMecAnn(adata, mec_score_topg, sigscaling=True, genescaling=False, addon=False)
    """

    adata = adata_input.copy()
    df = annToDataFrame( adata, genescaling = genescaling, layer = layer )    
    
    projected = projectMec(df, mec, glstcol = glstcol )
    projected.columns = [ str(t) for t in projected.columns ]
    if( sigscaling ):
        projected = scale(projected)
        
    if(addon):
        if( len(projected.columns.intersection(adata.obs.columns)) >0 ):
            warnings.warn('signature names already in adata.obs. cannot merge. If wish to remove in obs: \n \
                adata.obs = adata.obs[[t for t in adata.obs.columns if t not in mecnamedict.keys()]  \n or \
                    adata.obs = adata.obs[[t for t in adata.obs.columns if t[:6]!="score_" ]]]'
                )
        else:
            adata.obs = adata.obs.merge( projected , how = 'left', left_index = True, right_index = True)
        adataproj = adata.copy()
    else:
        adataproj = anndata.AnnData( projected, obs = adata.obs )
    return(adataproj)


def projectModuleAnn_aucell( adata, module, glstcol = 'TopGene20'):
    """ 
    Alternative function that projects scRNA data using top genes from MeCs and AUCell
    module has to be list mode.

    Parameters
    ----------
    adata: anndata.AnnData
        Input scanpy object for gene expression
    module: pd.DataFrame
        mec table with one column, two types of format accepted
        format 1, each row is a mec with genes , comma seperated.
    glstcol
        Used only when mec is the list format format 1, and glstcol is the column name to record comma separated gene list for each mec.

    Returns
    ----------
    anndata.AnnData
        adata with aucell score stored in extra columns starting with 'score_sig_' in adata.obs 

    
    """
    
    if(not glstcol): 
        glstcol = module.columns[0]
    for i in module.index:
        gene_list = module.loc[i,  glstcol ].split(',')
        use_gene_list = adata.var_names.intersection( gene_list )
        if(len(use_gene_list)>1):
            sc.tl.score_genes(adata, use_gene_list, score_name = 'score_sig_'+str(i) )
        else:
            adata.obs[ 'score_sig_'+str(i) ] = 0
    return(adata)

"""
### TODO. 
# Tutorial. mapper+annotator.
# Tutorial. differential signature.
# Tutorial functional extraction.
# metatime calling. (code to be wrapped in pipeline with a tutorial, that's it. )
# A class for projected data. this is per-cell class.
#init: score matrix, .obs features, cell location (adata.umap). function and category for features (mecs class.).
#fun:  plot single singature signature using mec name. plot all signature. plot all signature in
# 
#THEN, a seperate annotator function. takes projected class. takes adata input ( overclustering. )
# THEN, a separate differential signature comparison. 
# mecs class: print top gene, plot top gene, extract top tf, extract enrichr, extract lisa ranking, extract lisa-mec ranking plot. 
"""


"""
## class and functions for meta components
"""
import numpy as np
import pandas as pd
import os
import sys

from metatime import config

################  Special manual annotation of MetaTiME precomputed MeCs #############

def load_mecname( mode='table', 
                mecDIR= config.SCMECDIR
                ):
    """
    Load functional annotation for pre-computed tumor microenvironemnt MeCs

    Parameters
    ----------
    mode
        choose from ['mecnamedict', 'table', 'meciddict']
        load manual assigned name for easy understanding of assigned names
        from file: MeC_anno_name.txt under mecDIR.
        Required columns: `['MeC_id', 'Annotation', 'UseForCellStateAnno']` 
        Required seperator: tab
        Annotation column NA will be filtered.

    Returns
    ----------
    functional annotation in desided format

    """
    mecname = pd.read_table(os.path.join( mecDIR , 'MeC_anno_name.txt'))
    if(mode=='table'):
        mectable=mecname.copy()
        #mectable['color_level0'] = mectable['MajorLineage_level0'].apply(lambda t: config.level0colordict[t])
        return(mectable)        
    else:
        mask = (~mecname['Annotation'].isna())
        #mecname.loc[mask,'Annotation']=mecname.loc[mask,'MeC_id' ]
        mecnamedict = mecname[mask].set_index('MeC_id')[['Annotation']].to_dict()['Annotation']

    if(mode=='mecnamedict'):
        return(mecnamedict)

def getmecnamedict_ct( mectable, only_include_mecs_UseForCellStateAnno = True ):
    """ 
    Collect list of meta components to be used for cell state annotation. 

    Parameters
    ----------
    mectable
        must have `UseForCellStateAnno` column with 0 or 1. 1:used in cell state annotation. this is helpful to remove pan-cell cell state like general mitochondrial activity component.
    only_include_mecs_UseForCellStateAnno
        filter mecs used for cell state annotation. 
        Default: True 
    Returns
    ----------
    dict 
        a subsetted dictionary only containing MeCs used for cell state enrichment.

    """
    if(only_include_mecs_UseForCellStateAnno):
        mecnamedict = mectable[mectable['UseForCellStateAnno']==1].set_index('MeC_id')['Annotation'].to_dict()
    else:
        mecnamedict = mectable.set_index('MeC_id')['Annotation'].to_dict()
    return(mecnamedict)



class MetatimeMecs():
    """
    Class for MetaTiME Mecs.
    
    Parameters
    ----------
    modelpath
        directory of model files

    Attributes
    ----------
    mecscore
        pandas dataframe, z-weights  gene by component
    mectopg
        pandas dataframe, list for top gene only
    mecanno
        pandas dataframe, annotation for each mec. 

    """
    def __init__(self, mec_score, mec_topg, mec_anno ) :
        self.mec_score = mec_score
        self.mec_topg = mec_topg
        self.mec_anno = mec_anno

    @staticmethod
    def load_mec_precomputed(  mecDIR = config.SCMECDIR, ):
        """
        Load pre-computed Meta-component matrix and ordered list.
        Look for precomputed files in mecDIR

            MeC_allgene_average-weights.txt
            MeC_topgene.txt

        Parameters
        ----------
        mecDIR
            Path for model files

        Returns
        ----------
        MetatimeMecs
            Loaded MeCs    

        Examples 
        ----------
        >>> mecmodel = mecs.MetatimeMecs.load_mec_precomputed()

        """
        try:
            mec_score = pd.read_table( os.path.join( mecDIR , 'MeC_allgene_average-weights.txt') , index_col = 0)
            mec_topg = pd.read_table( os.path.join( mecDIR, 'MeC_topgene.txt') , index_col = 0)
            mec_anno = pd.read_table( os.path.join( mecDIR, 'MeC_anno.txt') , index_col = 0)
            # use 'MeC' name
            mec_score.columns = 'MeC_'+mec_score.columns
            return( MetatimeMecs( mec_score, mec_topg, mec_anno ) )

        except Exception as exception:
            raise Exception(f' Error loading model. check files in {mecDIR}. MeC_allgene_average-weights.txt, MeC_topgene.txt, MeC_anno.txt. {exception}')

    @property
    def feature(self) -> np.ndarray:
        """ get genes covered in the mec z-weight matrix """
        return( self.mec_score.index)
    
    @property
    def nmecs(self) -> np.float:
        """ get number of meta-components """
        return( self.mec_score.columns.shape[0])


    # interpretation functions



"""
# function for plotting after mapper
"""
import numpy as np 
import matplotlib.pyplot as plt
import importlib as imp
import pandas as pd
import scanpy as sc
import anndata
from adjustText import adjust_text

# set the colormap and centre the colorbar
import matplotlib.colors as colors

class MidpointNormalize(colors.Normalize):
    """
    Normalise the colorbar separately for above-midpoint and below-midpoint. Useful for visualizing the signature continuum with diverging colorbar. borrowed from [chris35wills](http://chris35wills.github.io/matplotlib_diverging_colorbar/)
    
    Parameters
    ----------
    colors.Normalize
    e.g. midpoint=0,vmin=vmin,vmax=vmax

    Examples 
    ----------
    # Pass kwargs to scatterplot
    >>> kwargs={'color_map':'RdBu_r', 'norm': MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }
    
    # note: 
    """
    
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y), np.isnan(value))


def plot_umap_proj( adata : anndata.AnnData, 
                    Xproj : pd.DataFrame, 
                    mecnamedict, 
                   use_MeC_name = True, 
                   figfile = None,
                   N_col = 4 ,
                   N_MECS = 100,
                   ):
    """
    Plotting function: plot the signature continuum score for each funcntional MeC in one large figure.

    Parameters
    ----------
    adata: anndata.AnnData. 
        scanpy object for scRNA data.
        Two formats are accepted.
        1. scanpy object with gene by cell expression, with extra mec columns in adata.obs. 
        When adata.var_names has less than 100 features, go with format1.
        2. scanpy object with only gene by MeC scores in adata.X. projected style format. 

    Xproj: pd.DataFrame
        Dataframe for projected scores


    Returns
    ----------
        matplotlib figure

    
    """
    if(len(adata.var_names)< N_MECS ):
        # adata is a projected format
        Xproj = adata.to_df()
    elif([t for t in Xproj.columns if t not in adata.obs.columns]):
        adata.obs = adata.obs.merge( Xproj,left_index=True,right_index=True,how='left')
    N_col = N_col
    N_row = int(Xproj.shape[1] /N_col)+1
    fig, axes=plt.subplots(nrows= N_row , ncols= N_col ,figsize=(N_col*3,N_row*2), sharex=True,sharey=True, num=86)

    for i in range( Xproj.shape[1] ):
    #for i in range( 8 ):
        i_row = int(i/N_col)
        i_col = i%N_col
        vmin = min( min(Xproj[ Xproj.columns[i] ].values) , -3 )
        vmax = max( max(Xproj[ Xproj.columns[i] ].values) , 3 )
        kwargs={'color_map':'RdBu_r', 'norm': MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }
        # 
        if( use_MeC_name ): 
            #title = str(i)+' ' +mecnamedict[ Xproj.columns[i] ]
            title = Xproj.columns[i].replace('score_','') +' ' +mecnamedict[ Xproj.columns[i] ]
        else:
            title = Xproj.columns[i].replace('score_','') 
        sc.pl.umap(adata, color= Xproj.columns[i], 
                        title = title,
                       show=False, ax = axes[i_row, i_col], **kwargs,)
        axes[i_row, i_col].set_xlabel('')
        axes[i_row, i_col].set_ylabel('')
    if(figfile ):
        plt.savefig( figfile , dpi=200 )
    return( axes )

########## plotting full panel for pdata format ############

## todo: group and rank the components

def plot_umap_proj_pdata( pdata, mecnamedict, 
                   use_MeC_name = True, 
                   figfile = None,
                   N_col = 4 ,
                   ):
    """
    Save all projection images in one large figure. input is pdata format. Deprecated
    """
    pdata = pdata[ : , list(mecnamedict.keys())]
    Xproj = pdata.to_df()
    N_col = N_col
    N_row = int(Xproj.shape[1] /N_col)+1
    fig, axes=plt.subplots(nrows= N_row , ncols= N_col ,figsize=(N_col*3,N_row*2), sharex=True,sharey=True, num=86)

    for i in range( Xproj.shape[1] ):
    #for i in range( 8 ):
        i_row = int(i/N_col)
        i_col = i%N_col
        vmin = min( min(Xproj[ Xproj.columns[i] ].values) , -3 )
        vmax = max( max(Xproj[ Xproj.columns[i] ].values) , 3 )
        kwargs={'color_map':'RdBu_r', 'norm': MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }
        # 
        if( use_MeC_name ): 
            #title = str(i)+' ' +mecnamedict[ Xproj.columns[i] ]
            title = Xproj.columns[i].replace('score_','') +' ' +mecnamedict[ Xproj.columns[i] ]
        else:
            title = Xproj.columns[i].replace('score_','') 
        sc.pl.umap(pdata, color= Xproj.columns[i], 
                        title = title,
                       show=False, ax = axes[i_row, i_col], **kwargs,)
        axes[i_row, i_col].set_xlabel('')
        axes[i_row, i_col].set_ylabel('')
    if(figfile ):
        fig.savefig( figfile , dpi=200 )
    return( fig )


def plot_umap_mec(pdata : anndata.AnnData,
                 meccol: str, 
                 mecnamedict: dict, 
                 use_MeC_name: bool = True, 
                 figfile: str = None,
                 figsize=(3,3)
                 ):

    """
    Plot signature continuum for a specific MeC. 

    Parameters
    ----------
    pdata: anndata.AnnData 
        scanpy object with projection score matrix in pdata.X
    meccol
        MeC id
    mecnamedict


    Returns
    ----------

    """
    Xproj = pdata.to_df()

    fig,ax =plt.subplots(1,figsize= figsize )

    # keep value same range
    vmin = min( min(Xproj[meccol] .values) , -3 )
    vmax = max( max(Xproj[meccol] .values) , 3 )
    kwargs={'color_map':'RdBu_r', 'norm': MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }

    # figure title
    if( use_MeC_name ): 
        title = meccol.replace('score_','') +' ' +mecnamedict[ meccol ]
    else:
        title = meccol.replace('score_','') 

    sc.pl.umap(pdata, color= meccol, title=title,
                show=False, ax=ax, **kwargs,)
    if(figfile ):
        plt.savefig( figfile , dpi=200 )        
    return( fig )




def plot_umap_mecproj_2condition( pdata, meccol, mecnamedict, 
                    cellscond1,cellscond2,
                   use_MeC_name = True, 
                   figfile = '../tmp.png',
                  
                   ):
    """
    
    Pplotting comparison panel for pdata format, beta version.

    Examples
    ----------
    `Xproj = pdata.to_df()
    use_MeC_name = True
    mec_col ='score_1'
    mec_name = mecnamedict[ mec_col ]
    condcol = 'response'
    # cells from 2 conditions
    allcellscond1= adata[adata.obs[ condcol ].isin(['Non-responder']), ].obs.index
    allcellscond2= adata[adata.obs[ condcol ].isin(['Responder']), ].obs.index

    plmapper.plot_umap_mecproj_2condition( pdata, meccol = 'score_0', mecnamedict = mecnamedict, 
                                        cellscond1 = allcellscond1, cellscond2 = allcellscond2, 
                                        use_MeC_name = True, )
                                     
    `
    """
    Xproj = pdata.to_df()

    if((len(cellscond1)==0) or(len(cellscond2)==0)):
        return(None)
    # plot panel
    fig, axes=plt.subplots(nrows= 1 , ncols= 2 ,figsize=(2*3,1*3), sharex=True,sharey=True, num=2)

    # keep value same range
    vmin = min( min(Xproj[meccol] .values) , -3 )
    vmax = max( max(Xproj[meccol] .values) , 3 )
    kwargs={'color_map':'RdBu_r', 'norm': MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }
    """
    # figure title
    if( use_MeC_name ): 
        title = meccol.replace('score_','') +' ' +mecnamedict[ meccol ]
    else:
        title = meccol.replace('score_','') 
        """
    # condition1,2
    sc.pl.umap(pdata[ pdata.obs.index.intersection( cellscond1)  ], color= meccol, 
                    title = 'cond1',
                show=False, ax = axes[0], **kwargs,)

    sc.pl.umap(pdata[ pdata.obs.index.intersection( cellscond2) ], color= meccol, 
                    title = 'cond2',
                show=False, ax = axes[1], **kwargs,)
    axes[0].set_xlabel('')
    axes[1].set_ylabel('')
    if(figfile ):
        plt.savefig( figfile , dpi=200 )        
    return( fig )





############# annotator plots #############


def gen_mpl_labels(
    adata, groupby, exclude=(), ax=None, adjust_kwargs=None, text_kwargs=None
):
    """ 
    Get locations of cluster median . Borrowed from scanpy github forum.
    """
    if adjust_kwargs is None:
        adjust_kwargs = {"text_from_points": False}
    if text_kwargs is None:
        text_kwargs = {}

    medians = {}

    for g, g_idx in adata.obs.groupby(groupby).groups.items():
        if g in exclude:
            continue
        medians[g] = np.median(adata[g_idx].obsm["X_umap"], axis=0)

    if ax is None:
        texts = [
            plt.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()
        ]
    else:
        texts = [ax.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()]

    adjust_text(texts, **adjust_kwargs)

def plot_annotation_on_data( 
    sdata, 
    COL='MetaTiME'  ,
    title = None, 
    fontsize = 8,
    MIN_CELL = 5,
    ):
    """ 
    Plot annotated cells with  non-overlapping fonts.
    Calls: gen_mpl_labels
    Occasionally, if sdata.uns['MetaTiME_overcluster_colors'] exists, color may not be updated. In that case, resetting color  `del pdata.uns['MetaTiME_overcluster_colors']`
    
    Parameters
	----------
	sdata : anndata.AnnData
        scanpy object for single cell , or scanpy object with projected signature. 
    COL : str
        column of cluster assignment in sdata.obs
    MIN_CELL: int
        Minimum number of cells to plot and mark

	Returns
	----------
        Matplotlib figure.

    Examples 
    ----------
    >>> plmapper.plot_annotation_on_data(pdata, title = 'MetaTiME')
    
    """
    #
    if not title:
        title = COL

    if( MIN_CELL >0 ):
        groupcounts = sdata.obs.groupby( COL ).count()
        groupcounts = groupcounts[groupcounts.columns[0]]
        group_with_good_counts = groupcounts[groupcounts>= MIN_CELL ].index.tolist()
        sdata = sdata[ sdata.obs[ COL ].isin( group_with_good_counts ) ]

    with plt.rc_context({"figure.figsize": (6, 6), "figure.dpi": 300, "figure.frameon": False}):
        #ax = sc.pl.umap(pdata, color="MetaTiME_overcluster", show=False, legend_loc=None, frameon=False, size=30)
        ax = sc.pl.umap(sdata, color= COL , show=False, legend_loc=None, add_outline=False, 
                #legend_loc='on data',legend_fontsize=6, legend_fontoutline=2,
                title= title, 
                palette=plt.cycler("color",plt.cm.tab20(np.linspace(0,1,20))), 
                #palette=plt.cycler("color",plt.cm.Set1(np.linspace(0,1,9))), 
                    )
        gen_mpl_labels(
            sdata,
            COL,
            exclude=("None",),  
            ax=ax,
            adjust_kwargs=dict(arrowprops=dict(arrowstyle='-', color='black')),
            text_kwargs=dict(fontsize= fontsize ,weight='bold'),
        )
        fig = ax.get_figure()
        fig.tight_layout()
        #plt.show()
        return( fig,ax )
        #plt.savefig('MetaTiME_overcluster.pdf', dpi=200)


############## snippets ##########
"""
# save all projection images in single figures.

for i in range( Xproj.shape[1] ):
#for i in range( 2 ):
    vmin = min( min(Xproj[ Xproj.columns[i] ].values) , -2 )
    vmax = max( max(Xproj[ Xproj.columns[i] ].values) , 2 )
    kwargs={'color_map':'RdBu_r', 'norm': plmapper.MidpointNormalize(midpoint=0,vmin=vmin,vmax=vmax) }
    ax = sc.pl.umap(adata, color= Xproj.columns[i], **kwargs,
                    title = mecnamedict[ Xproj.columns[i] ],
                   show=False )
#    plt.savefig('../analysis/20211001_scICA/BRCA_Alex/BRCA_Alex.rawumap.mec_'+str(keyi)+'.png')
    plt.close()
    
    """



"""
Automatic annotator for tumor single cell data. 
# 
# """
### MetaTiME annotator
import pandas as pd
import anndata
from metatime import config
import scanpy as sc
from typing import Optional, Union
from metatime import mecs
import warnings



 
########### top1 cell state annotator #######
def annotator( dat: Union[pd.DataFrame, anndata.AnnData],
              mecnamedict: dict, 
              gcol ='overcluster', 
              #MINCELL = 5 
              ):  
    """ 
    Annotator marking top1st enriched cell states based on MeC scores.

    Parameters
    ----------
    dat
        dataframe with gene by cell scores and one column gcol (grouping column) e.g. leiden class
        or,
        anndata with gene by cell scores, and one column gcol in dat.obs 
        Malignant cells shall be removed to keep tumor microenvironmental cells only, such as immune cells, fibroblasts, endothelial cells
    mecnamedict
        for renaming dat columns into functional names
        can be loaded from pre-computed tumor MeC functional annotation
    gcol
        for grouping cells. e.g. a column for overclustered cluster assignment.

    Returns
    ----------
    projmat: pd.DataFrame
         Dataframe with mec projection scores and newly added predicted label column 'MetaTiME_'+gcol
    gpred
        median score of each gcol group
    gpreddict
        dictionary how is each gcol mapped to a label
    
    Examples
    ----------
    >>> projmat, gpred, gpreddict = annotator( projmat ,  mecnamedict)
    
    """
    if (isinstance( dat , anndata.AnnData)):
        # subselecting columns
        pdata = dat.copy()
        Xproj = pdata.to_df()
        #projmat = Xproj[Xproj.columns.intersection(config.scorectcols)].join(pdata.obs[[ gcol ]])
        projmat = Xproj.join(pdata.obs[[ gcol ]])
    else:
        projmat = dat.copy()

    projmat = projmat[ list( mecnamedict.keys() ) + [ gcol ]]

    # mean or median score for each group
    gpred = projmat.groupby( gcol ).median() # gpred: leiden cluster by mec.
    # top 1 score for each group, with filtering
    gpreddict = projmat.groupby( gcol ).median().idxmax(axis=1).apply(lambda t: mecnamedict[t])# cluster max quantile 75% larger than zscore-2 
    pass_maxmedian = (projmat.groupby( gcol ).quantile(q=0.7) >1 ).max(axis=1) # given median is max, top 30%cells shall pass zscore 1. Otherwise this may be a noisy group, setting to 'Others'
    gpreddict[ ~ pass_maxmedian] = 'Others'
    # generate top1 prediction
    projmat['MetaTiME_'+gcol] = projmat[ gcol ].apply(lambda t: gpreddict[t])
    
    # add to anndata
    if (isinstance( dat , anndata.AnnData)):
        if('MetaTiME_'+gcol in pdata.obs.columns):
            del pdata.obs['MetaTiME_'+gcol]
        pdata.obs['MetaTiME_'+gcol] = projmat['MetaTiME_'+gcol]
        pdata.obs['MetaTiME_'+gcol]=pd.Categorical(pdata.obs['MetaTiME_'+gcol])
        dat = pdata
    else:
        dat = projmat
    
    return( dat, gpred, gpreddict )


def overcluster(adata: anndata.AnnData,
                resolution : float=8, 
                random_state: int= 0, 
                clustercol :str = 'overcluster'):

    """
    Overcluster single cell data to get cluster level cell state annotation

    Parameters
    ----------
    adata
        scanpy object with adata.uns['neighbors'] computed.
        if adata.obsm['X_umap'] does not exist, recomputes umap coordinates.
        otherwise, keep the umap coordinates
    resolution
        clustering resolution
    random_state
        clustering random state
    clustercol:
        key to add to adata.obs that records cluster assignment

    Returns
    ----------
        scanpy object with clustering results.

    """
    sc.tl.leiden(adata, resolution=resolution, key_added = clustercol, random_state=random_state)
    if 'X_umap' not in adata.obsm.keys():
        sc.tl.umap(adata)
    return(adata)

def pdataToTable( pdata: anndata.AnnData, 
                    mectable: pd.DataFrame, 
                    gcol : str= 'overcluster'):

    """
    Convert projected scores to two simple pandas dataframes

    Parameters
    ----------
    pdata
        anndata with gene by mec scores, and one column gcol in pdata.obs 
    mectable
        for renaming dat columns into functional names
        can be loaded from pre-computed tumor MeC functional annotation
        Required columns: `['MeC_id', 'Annotation', 'UseForCellTypeAnno']`  
    gcol
        a column in pdata.obs for grouping cells. a column for overclustered cluster assignment.

    Returns
    ----------
    tuple
        projmat : a pandas dataframe with mec scores and a column for grouping cells. useful for annotating cell states. 
        mecscores: a pandas dataframe for per-cell mec scores. columns use functional annotation of mec ids


    """
    # projmat
    projmat = pdata.obs.join( pdata.to_df() )
    allmecids = mectable['MeC_id'].values.tolist()
    projmat = projmat[ allmecids  + [ gcol ] ]

    # mecscores
    mecscores = pdata.to_df()
    mecctnamedict = mecs.getmecnamedict_ct( mectable )
    newcols = []
    for t in projmat.columns:
        if(t in mecctnamedict.keys() ):
            newcols.append( mecctnamedict[t]  )
        else:
            newcols.append(t)
    mecscores = projmat.copy()
    mecscores.columns = newcols
    
    return(projmat, mecscores)


def saveToAdata( adata : anndata.AnnData, 
                projmat : pd.DataFrame, 
                gcol: str = 'overcluster',
                ANNOTATION_ONLY : bool= False
                ):
    """
    Save annotation to adata. 

    Parameters
    ----------
    adata: anndata.AnnData
        scRNA scanpy object
    projmat: pd.DataFrame
        Dataframe with mec projection scores and newly added predicted label column 'MetaTiME_'+gcol
    gcol
        A column in projmat for grouping cells. Typically a column for overclustered cluster assignment.
    ANNOTATION_ONLY
        Whether to only add cluster-wise annotation to anndata or Also append scores to adata.obs

    Returns
    ----------
    anndata.AnnData
        adata with per-cluster annotation column in adata.obs[[gcol]], and scores appended to adata.obs if ANNOTATION_ONLY==True

    """
    exist_cols = projmat.columns.intersection(adata.obs.columns)
    if( len(exist_cols ) >0 ):
            warnings.warn('Columns in projmat already overlap columns in adata.obs. Overwriting adata.obs. '
                )
            adata.obs = adata.obs[[ t for t in adata.obs.columns if t not in exist_cols] ]
    
    if( ANNOTATION_ONLY ):
        adata.obs['MetaTiME_'+gcol] = projmat['MetaTiME_'+gcol]
    else:
        adata.obs = adata.obs.merge( projmat , how = 'left', left_index = True, right_index = True)
    return( adata )
        

def saveToPdata( pdata : anndata.AnnData, 
                adata: anndata.AnnData,
                projmat : pd.DataFrame, 
                gcol: str = 'overcluster',
                BORROW_ADATA_EMBEDDING=True,
                ):
    """
    Save annotation to pdata.
    Borrow embedding from adata for easy visualization, including adata.obsm['X_pca'], adata.obsm['X_umap'], adata.obsm['X_pca_harmony']

    Parameters
    ----------
    pdata : anndata.AnnData
        scanpy object for per-cell projected score. 
    adata: anndata.AnnData
        scRNA scanpy object
    projmat: pd.DataFrame
        Dataframe with mec projection scores and newly added predicted label column 'MetaTiME_'+gcol
    gcol
        A column in projmat for grouping cells. Typically a column for overclustered cluster assignment.
    BORROW_ADATA_EMBEDDING
        Whether to borrow pca and umap embeddings from adata to write in pdata. For easy visualization.

    Returns
    ----------
    anndata.AnnData
        pdata with per-cluster annotation column in pdata.obs[[gcol]]. 

    """
    if(BORROW_ADATA_EMBEDDING):
        pdata.obsm['X_pca'] = adata.obsm['X_pca']
        pdata.obsm['X_umap'] = adata.obsm['X_umap']
        if('X_pca_harmony' in adata.obsm):
            pdata.obsm['X_pca_harmony'] = adata.obsm['X_pca_harmony']

    pdata.obs[gcol] = projmat[[gcol]]
    pdata.obs['MetaTiME_'+gcol] = projmat['MetaTiME_'+gcol]
    return( pdata )





import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='metatime',
    version='1.2.2',
    author='Yi Zhang',
    author_email='wingsyiz@gmail.com',
	description='Beta MetaTiME: annotate TME scRNA cell states',
    long_description  = long_description,
    long_description_content_type="text/markdown",
	url='https://github.com/yi-zhang/MetaTiME.git',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    packages=['metatime'],
	package_data = {
		'metatime': ['pretrained/mec/*txt']
		},
    python_requires='>=3.6',
    install_requires=['pandas','scanpy','anndata','matplotlib','adjustText', 'leidenalg', 'harmonypy', 'scipy','seaborn'],
)



"""
## calculate differential mec signature comparing two conditions
"""
import numpy as np
import matplotlib.pyplot as plt
import importlib as imp
from typing import Optional, Union
import pandas as pd
import scanpy as sc
import seaborn as sns
import anndata
import scipy.stats as ss

def _get_score_pd( adata, meccols ):
    """ 
    Obtain scores from scanpy object.
    compatible with:
        scanpy object with MetaTiME projection score in .X .
        or,
        scanpy object with MetaTiME projection score saved in .obs[meccols]
    """
    if meccols[0] in adata.var_names:
        x = adata.to_df()[meccols]
    elif( meccols[0] in adata.obs.columns):
        x = adata.obs[ meccols ]
    else:
        return(None)
    return(x)

def count_cell_condition( obs,  countbycols : list):
    """
    Counting cells by condition 

    Params
    ----------
    obs: adata.obs,
        metadata dataframe
    countbycols: list, has to be columns in obs.

    Examples
    ----------
    condcol = 'treatment' 
    clustercol = 'MetaTiME' 
    samplecol = 'patient'
    counts = count_cell_condition( adata,obs, countbycols  = [condcol, clustercol, samplecol] )

    """
    cellcts = obs.groupby( countbycols ).count()['MetaTiME_overcluster'].fillna(0).unstack()
    return(cellcts)

def enrich_mec_cluster(  labeltype, adata, labelcol , meccols , mecnamedict ):
    """ 
    For one cluster, Output which MeC is enriched.
    criterion: MeC score cluster-mean passes zscore 1 (1-standard deviation)
    
    Parameters
    ----------
    adata: anndata.AnnData.
        compatible with:
        scanpy object with MetaTiME projection score in .X .
        or,
        scanpy object with MetaTiME projection score saved in .obs[meccols]

    """
    if meccols[0] in adata.var_names:
        x = adata[adata.obs[ labelcol ]==labeltype].to_df()[meccols]
    elif( meccols[0] in adata.obs.columns):
        x = adata.obs.loc[ adata.obs[ labelcol ]==labeltype,meccols ]
    else:
        return(False)
    tscores = x.mean(axis=0).sort_values(ascending=False)
    #tscores = x.median(axis=0).sort_values(ascending=False)
    tscores = tscores[tscores>=1]#  [:2] # top 2 pass 1std
    tscores = tscores.rename(mecnamedict)
    topmecs = tscores.index.tolist()
    
    return(tscores.to_dict())

def enrich_mec( adata , labelcol , mecnamedict ):
    """ 
    Determine enriched MeCs per group or cluster.

    Parameters
    ----------
    adata: anndata.AnnData.
        compatible with:
        scanpy object with MetaTiME projection score in .X .
        or,
        scanpy object with MetaTiME projection score saved in .obs[meccols]

    labelcol: str
        The column name of grouping cells
    mecnamedict
        for renaming dat columns into functional names
        can be loaded from pre-computed tumor MeC functional annotation

    Returns
    ----------
    pd.DataFrame
        A data frame with each cluster in a row, and a column 'topmec' listing MeCs with significant enrichment as candidate for testing differential signature.

    Example
    ----------
    `
        cluster_mec_enriched = dmec.enrich_mec( labeltypes, adata=adata, labelcol =labelcol, meccols = meccols, mecnamedict=mecnamedict)
    `
    """
    # columns for meta components
    meccols = list(mecnamedict.keys())
    # labeltypes: categories in the cell grouping column 
    labeltypes = adata.obs[labelcol].drop_duplicates().tolist()
    
    res1 = pd.DataFrame( index=labeltypes)
    res={}
    for labeltype in labeltypes:
        topmec = enrich_mec_cluster( labeltype ,  adata=adata, labelcol =labelcol, meccols = meccols , mecnamedict =mecnamedict) 
        res.update( {labeltype: topmec})
        res1.loc[labeltype,'topmec']= ','.join(topmec)
    return(res1)

def cohend( x1, x2 ):
    """ t  test effect size """
    cohens_d = (np.mean(x1) - np.mean(x2)) / (np.sqrt((np.std(x1) ** 2 + np.std(x2) ** 2) / 2))
    return(cohens_d)

def test_sig_smallcluster(Xproj, meci, cellscond1, cellscond2):
    """ 
    testing signature difference at each cell group, for certain mec signatures in meci.
    """
    
    x1 = Xproj.loc[cellscond1, meci].values
    x2 = Xproj.loc[cellscond2, meci].values

    #cut=2
    #y1=x1[x1>=cut]
    #y2=x2[x2>=cut]
    #[s,p]=ss.ranksums(x2, x1)
    [s,p]=ss.ttest_ind(x2, x1)
    #u,p = ss.mannwhitneyu(x1,x2)
    #logfc = np.log2((np.mean(y2)/len(x2))/(np.mean(y1)/len(x1)))# consider cell number differences
    #logfc = np.log2((np.median(y2)+1)/(np.median(y1)+1))# consider cell number differences
    #logfc = np.log2( (np.median(x2) +1 )/(np.median(x1) +1 ) )
    effect = (np.mean(x2)  - np.mean(x1) )  
    
    #logfc = cohend( x1, x2)
    #logfc = np.log2((np.mean(x2))/(np.mean(x1)))
    meanb=np.mean(x2); meana=np.mean(x1)
    #logctfc = np.log2((len(y2)/len(x2))/(len(y1)/len(x1)))# consider cell number differences
    #logctfc = np.log2((np.sum(y2)/len(x2))/(np.sum(y1)/len(x1)))# consider cell number differences
    #logctfc = np.log2( (np.sum(x2)/N2 +1)/(np.sum(x1)/N1 +1))# consider cell number differences
    effect1 = effect

    logp=-np.log10(p)
    return( logp, effect,effect1,meanb,meana )
### plot enriched 1st mec

def plot_top1mec( adata, mec_enriched , legend_loc ='on data'):
    mec_enriched['top1mec'] = mec_enriched['topmec'].str.split(',').str.get(0)
    adata.obs = adata.obs.merge( mec_enriched[['top1mec']], left_on = 'celltype_subset', right_index=True, how='left')
    sc.pl.umap(adata, color = ['top1mec'], legend_fontsize='5', legend_loc= legend_loc , alpha=0.4 )


def dmec( adata, 
          pdata,
          allcellscond1,
          allcellscond2,
          clustercol: str  ,
          cluster_mec_enriched: pd.DataFrame, 
          mecnamedict,
          test_clusters = None ,):
    """ 
    Differential signature analysis

    Parameters
    ----------
    adata
        scRNA object
    pdata
        Projection score table, cell by signature. 
    allcellscond1
        list of barcodes of cells from condition 1
    allcellscond2
        list of barcodes of cells from condition 2
    clustercol
        cell cluter column in adata.obs 
    cluster_mec_enriched
        what mec to test in each cluster in test_clusters
    mecnamedict
        for renaming dat columns into functional names
        can be loaded from pre-computed tumor MeC functional annotation
    test_clusters
        Which cell clusters to test based on adata.obs[clutsercol]

    Returns
    ----------
    diffmec
        dictonary containing per-cluster differential testing results, good for summarizing and generating plots
    diffmecsig
        dictonary containing per-cluster differential testing results filtered to contain significant ones
    diffmec_full
        dictonary containing per-cluster differential testing results with more details

    """
    if(not test_clusters or test_clusters == 'all'):
        test_clusters =adata.obs[clustercol].drop_duplicates().tolist()
    diffmec_full={}
    diffmec={}
    diffmecsig={}
    for test_cluster in test_clusters:
        cell_clusteri = adata[adata.obs[ clustercol ]== test_cluster ].obs.index
        cellscond1 = allcellscond1.intersection(cell_clusteri)
        cellscond2 = allcellscond2.intersection(cell_clusteri)
        if((len(cellscond1)==0) or(len(cellscond2)==0)):
            continue
        # signatures that has high (>1std) score for this cluster
        clusters_strong_sig1 = cluster_mec_enriched.loc[test_cluster,'topmec'].split(',')
        if(len(clusters_strong_sig1)==0 or clusters_strong_sig1[0]==''):
            continue


        Xproj = pdata.to_df()
        res = pd.DataFrame(index=Xproj.columns)
        #N2 = len(allcellscond2);N1 = len(allcellscond1)
        for meci in Xproj.columns:
            t = test_sig_smallcluster(Xproj, meci, cellscond1, cellscond2 )
            res.loc[meci, '-logp']= t[0]
            res.loc[meci, 'effect']= t[1]
            res.loc[meci, 'effect1']= t[2]
            res.loc[meci, 'meanb']= t[3]
            res.loc[meci, 'meana']= t[4]

        diff_mec = res.sort_values(ascending=False,by='effect')
        diff_mec['scorecol'] = diff_mec.index
        diff_mec = diff_mec.rename(index=mecnamedict)

        diff_mec_matchedmec = diff_mec.loc[clusters_strong_sig1]
        diff_mec_matchedmec_sig = diff_mec_matchedmec[diff_mec_matchedmec['-logp']>=2].sort_values(ascending=False, by='effect')
        tmp = {'diff_mec_matchedmec_sig': diff_mec_matchedmec_sig,
               'diff_mec': diff_mec,
               'full_diff': res}
        diffmec_full.update({test_cluster : tmp })
        if(len(diff_mec_matchedmec_sig)>0):
            diffmecsig.update({ test_cluster: diff_mec_matchedmec_sig })
        if(len(diff_mec)>0):
            diffmec.update({ test_cluster: diff_mec_matchedmec })
    return( diffmec, diffmecsig, diffmec_full)


def plotdata_diffmec( diffmec, mecnamedict):
    """
    """
    import seaborn as sns
    import matplotlib.pyplot as plt
    from plmapper import MidpointNormalize
    mat_p = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_eff = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_b = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_a = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    for i in diffmec.keys():
        #print(i)
        mat_p.loc[i, diffmec[i].index] = diffmec[i]['-logp'].values
        mat_b.loc[i, diffmec[i].index] = diffmec[i]['meanb'].values
        mat_a.loc[i, diffmec[i].index] = diffmec[i]['meana'].values
        mat_eff.loc[i, diffmec[i].index] = diffmec[i]['logctfc'].values
    cluster_order = mat_p.max(axis=1).sort_values(ascending=False).index
    
    mec_order = mat_p.max(axis=0).sort_values(ascending=False).index
    tmp = pd.DataFrame(mat_p.loc[mat_p.index.intersection(cluster_order), mat_p.columns.intersection( mec_order) ].stack().rename('-logp'))
    #tmp = pd.DataFrame(mat_p.loc[cluster_order, mec_order].stack().rename('-logp'))
    tmp=tmp.merge( pd.DataFrame(mat_eff.stack().rename('effect')) , left_index=True,right_index=True, how='left')
    tmp=tmp.merge( pd.DataFrame(mat_b.stack().rename('meanb') ) , left_index=True,right_index=True, how='left')
    tmp=tmp.merge( pd.DataFrame(mat_a.stack().rename('meana') ) , left_index=True,right_index=True, how='left')
    tmp = tmp.reset_index()
    return(tmp)

def plot_diffmec( diffmec, mecnamedict):
    """
    """
    import seaborn as sns
    import matplotlib.pyplot as plt
    
    from metatime import plmapper
    from metatime.plmapper import MidpointNormalize
    mat_p = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_eff = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_b = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    mat_a = pd.DataFrame(index = diffmec.keys(), columns = mecnamedict.values())
    for i in diffmec.keys():
        #print(i)
        mat_p.loc[i, diffmec[i].index] = diffmec[i]['-logp'].values
        mat_b.loc[i, diffmec[i].index] = diffmec[i]['meanb'].values
        mat_a.loc[i, diffmec[i].index] = diffmec[i]['meana'].values
        mat_eff.loc[i, diffmec[i].index] = diffmec[i]['effect'].values
    cluster_order = mat_p.max(axis=1).sort_values(ascending=False).index
    
    mec_order = mat_p.max(axis=0).sort_values(ascending=False).index
    tmp = pd.DataFrame(mat_p.loc[mat_p.index.intersection(cluster_order), mat_p.columns.intersection( mec_order) ].stack().rename('-logp'))
    #tmp = pd.DataFrame(mat_p.loc[cluster_order, mec_order].stack().rename('-logp'))
    tmp=tmp.merge( pd.DataFrame(mat_eff.stack().rename('effect')) , left_index=True,right_index=True, how='left')
    tmp=tmp.merge( pd.DataFrame(mat_b.stack().rename('meanb') ) , left_index=True,right_index=True, how='left')
    tmp=tmp.merge( pd.DataFrame(mat_a.stack().rename('meana') ) , left_index=True,right_index=True, how='left')
    tmp = tmp.reset_index()

    sns.set_theme(style="whitegrid")
    kwargs={'hue_norm': MidpointNormalize(midpoint=0,vmin=min(tmp['effect']), vmax = max(tmp['effect']) ) ,
            }
    g = sns.relplot(data = tmp, 
                x = 'level_0', y='level_1', hue = 'effect', size = '-logp',
                palette="RdBu_r",
                #hue_norm=(-1, 1), 
                edgecolor=".7",
        height=8, aspect = 1.2, 
        #width=8, 
        sizes=(20, 100), size_norm=(-.2, .8),
                **kwargs
    )
    g.set(xlabel="", ylabel="", aspect="equal")
    g.despine(left=True, bottom=True)
    g.ax.margins(.02)
    for label in g.ax.get_xticklabels():
        label.set_rotation(90)
    for artist in g.legend.legendHandles:
        artist.set_edgecolor(".7")
    return(g)


def topdiff_df(diffmec,
             CUT_LOGPPOS = 1.3,
             CUT_EFFECTPOS = 1,
             CUT_EFFECTNEG = -1,
             FCCOL = 'effect'
             ):
    """
    Organize per-cluster differential signature results into tables

    Parameters
    ----------
    diffmec
        differential expression dict generated by dmec.dmec
    CUT_LOGPPOS 
        negative log p value cutoff to determin significance, default: 1.3 (#p=0.5)
    CUT_EFFECTPOS
        positive effect size to cut . default: 1 ( 1 std shift)
    CUT_EFFECTNEG
        negative effect size to cut . default: -1 ( 1 std shift)
    FCCOL
        the 'effect' to use (fold change)

    Example
    ----------
    `diff, diff_sig = dmec.topdiff_df(diffmec)`

    """
    tlst = []
    for cluster1 in diffmec.keys():
        df = diffmec[ cluster1 ].copy()
        df['cluster1']=cluster1
        df['mec']=df.index
        df.index= df.index + '@' + cluster1
        # MHCI@Mac_IL1B; Mac_IL1B (not Mac_IL1B@Mac_IL1B)
        df.index=df.index.str.replace( cluster1+'@'+cluster1 , cluster1 )
        tlst.append(df)
    dat = pd.concat(tlst)
    sigpos = dat[( dat[ FCCOL ]>= CUT_EFFECTPOS ) & ( dat['-logp']>= CUT_LOGPPOS ) & ( (dat['meanb']>=1) )]
    #sigpos = dat[( dat[ FCCOL ]>= CUT_EFFECTPOS ) & ( dat['-logp']>= CUT_LOGPPOS ) ]
    signeg = dat[( dat[ FCCOL ]<= CUT_EFFECTNEG ) & ( dat['-logp']>= CUT_LOGPPOS ) & ( (dat['meana']>=1) )]
    #signeg = dat[( dat[ FCCOL ]<= CUT_EFFECTNEG ) & ( dat['-logp']>= CUT_LOGPPOS ) ]
    markpos = sigpos.sort_values(ascending=False, by= FCCOL ).index
    markneg = signeg.sort_values(ascending=True, by= FCCOL ).index
    dat_sig = dat.loc[markpos.tolist() + markneg.tolist()].sort_values(ascending=False, by ='-logp')
    return( dat, dat_sig )


def plot_topdiff( diffmec, fontsize = 5 ):
    """
    Parameters
    ----------
    diffmec
        dictonary containing per-cluster differential testing results
    fontsize
        fontsize on the differential scatter plot

    Example
    ----------
    `sns.set_theme(style='white')
    fig_topdiff = dmec.plot_topdiff(diffmec, fontsize=6)
    `
    """
    CUT_LOGPPOS = 1.3 #p=0.5
    CUT_EFFECTPOS = 1 # 1 std shift
    CUT_EFFECTNEG = -1 # 1 std shift
    #FCCOL = 'logctft'
    FCCOL = 'effect'
    tlst = []
    for cluster1 in diffmec.keys():
        df = diffmec[ cluster1 ].copy()
        df['cluster1']=cluster1
        df['mec']=df.index
        df.index= df.index + '@' + cluster1
        # MHCI@Mac_IL1B; Mac_IL1B (not Mac_IL1B@Mac_IL1B)
        df.index=df.index.str.replace( cluster1+'@'+cluster1 , cluster1 )
        tlst.append(df)
    dat = pd.concat(tlst)

    fig,ax = plt.subplots(1, figsize=(6,5))
    sns.set_theme(style="white")
    sns.scatterplot(data=dat, x= FCCOL ,y='-logp', color='gray',legend=False,alpha=0.3)
    sigpos = dat[( dat[ FCCOL ]>= CUT_EFFECTPOS ) & ( dat['-logp']>= CUT_LOGPPOS ) & ( (dat['meanb']>=1) )]
    signeg = dat[( dat[ FCCOL ]<= CUT_EFFECTNEG ) & ( dat['-logp']>= CUT_LOGPPOS ) & ( (dat['meana']>=1) )]
    markpos = sigpos.sort_values(ascending=False, by= FCCOL ).index
    markneg = signeg.sort_values(ascending=True, by= FCCOL ).index

    # tuning figure
    lim = np.max([abs(t) for t in ax.get_xlim()])
    ax.set_xlim([-lim,lim])
    #ax.set_ylim([-30, ax.get_ylim()[1]])
    ax.axvline( CUT_EFFECTPOS ,linestyle='--', color='gray', alpha=0.4)
    ax.axvline( CUT_EFFECTNEG,linestyle='--', color='gray', alpha=0.4)
    ax.axvline( 0,linestyle='--', color='gray', alpha=0.6)
    ax.axhline( CUT_LOGPPOS ,linestyle='--', color='gray', alpha=0.4)
    # color
    sns.scatterplot(data=sigpos, x= FCCOL ,y='-logp', color='brown', size='meanb',legend=False, ax=ax)
    sns.scatterplot(data=signeg, x= FCCOL ,y='-logp', color='steelblue', size='meana',legend=False, ax=ax)
    # text
    texts=[]
    for meci in markpos:
        texts.append( plt.text( sigpos.loc[meci, FCCOL ],  sigpos.loc[meci,'-logp'], meci.replace('@','@') , fontsize= fontsize, color='red'))
    for meci in markneg:
        texts.append( plt.text( signeg.loc[meci, FCCOL ],  signeg.loc[meci,'-logp'], meci.replace('@','@') , fontsize= fontsize, color='blue'))

    from adjustText import adjust_text
    adjust_text(texts, arrowprops=dict(arrowstyle='-', color='gray'), autoalign='xy')
    ax.set_xlabel('Effect')
    return( fig, ax )

